'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

/* eslint no-unused-vars: "off", @typescript-eslint/no-explicit-any: "off", no-use-before-define: "off" */
(function (Directions) {
    Directions["LEFT"] = "left";
    Directions["RIGHT"] = "right";
    Directions["UP"] = "up";
    Directions["DOWN"] = "down";
    Directions["ENTER"] = "enter";
    Directions["UNSPECIFIED"] = "*";
})(exports.Directions || (exports.Directions = {}));
(function (Orientations) {
    Orientations["VERTICAL"] = "vertical";
    Orientations["HORIZONTAL"] = "horizontal";
})(exports.Orientations || (exports.Orientations = {}));

var KeyCodes = {
    4: exports.Directions.LEFT,
    21: exports.Directions.LEFT,
    37: exports.Directions.LEFT,
    214: exports.Directions.LEFT,
    205: exports.Directions.LEFT,
    218: exports.Directions.LEFT,
    5: exports.Directions.RIGHT,
    22: exports.Directions.RIGHT,
    39: exports.Directions.RIGHT,
    213: exports.Directions.RIGHT,
    206: exports.Directions.RIGHT,
    217: exports.Directions.RIGHT,
    29460: exports.Directions.UP,
    19: exports.Directions.UP,
    38: exports.Directions.UP,
    211: exports.Directions.UP,
    203: exports.Directions.UP,
    215: exports.Directions.UP,
    29461: exports.Directions.DOWN,
    20: exports.Directions.DOWN,
    40: exports.Directions.DOWN,
    212: exports.Directions.DOWN,
    204: exports.Directions.DOWN,
    216: exports.Directions.DOWN,
    29443: exports.Directions.ENTER,
    13: exports.Directions.ENTER,
    67: exports.Directions.ENTER,
    32: exports.Directions.ENTER,
    23: exports.Directions.ENTER,
    195: exports.Directions.ENTER
};

/**
 * Given an array of values and a goal, return the value from values which is closest to the goal.
 *
 * @param {number[]} values
 * @param {number} goal
 */
var closestIndex = function (values, goal) { return values.reduce(function (prev, curr) {
    return (Math.abs(curr - goal) < Math.abs(prev - goal) ? curr : prev);
}); };
/**
 * Checks if a given node is focusable.
 *
 * @param {object} node - node to check against focusability
 */
var isNodeFocusable = function (node) {
    if (!node) {
        return false;
    }
    return node.isFocusable != null ? node.isFocusable : !!node.selectAction;
};
/**
 * Given a keyCode, lookup and return the direction from the keycodes mapping file.
 *
 * @param {number} keyCode - key code for which corresponding direction is searched
 */
var getDirectionForKeyCode = function (keyCode) {
    return KeyCodes[keyCode];
};
/**
 * Given an orientation and a direction, do they match?
 *
 * I.e an orientation `horizontal` and direction `left` or `right` is considered matching.
 *
 * Direction CAN be passed as `*` (wildcard). this will always return true.
 *
 * @param {string} orientation - orientation to match with the direction
 * @param {string} direction - direction to match with the orientation
 */
var isDirectionAndOrientationMatching = function (orientation, direction) {
    if (!orientation || !direction) {
        return false;
    }
    var validOrientation = toValidOrientation(orientation);
    var validDirection = toValidDirection(direction);
    return ((validDirection === exports.Directions.UNSPECIFIED) ||
        (validOrientation === exports.Orientations.VERTICAL && (validDirection === exports.Directions.UP || validDirection === exports.Directions.DOWN)) ||
        (validOrientation === exports.Orientations.HORIZONTAL && (validDirection === exports.Directions.LEFT || validDirection === exports.Directions.RIGHT)));
};
/**
 * Returns a child from the given node whose indexRange encompasses the given index.
 *
 * @param {object} node - node, which children index ranges will be examined
 * @param {number} index - examined index value
 */
var findChildWithMatchingIndexRange = function (node, index) {
    if (!node || !node.children) {
        return undefined;
    }
    for (var _i = 0, _a = Object.keys(node.children); _i < _a.length; _i++) {
        var childId = _a[_i];
        var child = node.children[childId];
        if (child.indexRange && (child.indexRange[0] <= index && child.indexRange[1] >= index)) {
            return child;
        }
    }
    return undefined;
};
/**
 * Returns a child from the given node whose index matches the given index.
 *
 * @param {object} node - node, which children indices will be examined
 * @param {number} index - searched index value
 */
var findChildWithIndex = function (node, index) {
    if (!node || !node.children) {
        return undefined;
    }
    for (var _i = 0, _a = Object.keys(node.children); _i < _a.length; _i++) {
        var childId = _a[_i];
        var child = node.children[childId];
        if (child.index === index) {
            return child;
        }
    }
    return undefined;
};
/**
 * Returns a child from the given node whose index is numerically closest to the given index.
 *
 * If an indexRange is provided, first check if the node's activeChild is inside
 * the indexRange. If it is, return the activeChild node instead
 *
 * @param {object} node
 * @param {number} index
 * @param {number[]} indexRange
 */
var findChildWithClosestIndex = function (node, index, indexRange) {
    if (!node || !node.children) {
        return undefined;
    }
    // if we have an indexRange, and the nodes active child is inside that index range,
    // just return the active child
    var activeChild = node.children[node.activeChild];
    if (indexRange && activeChild && activeChild.index >= indexRange[0] && activeChild.index <= indexRange[1] && isNodeFocusable(activeChild)) {
        return activeChild;
    }
    var indexes = Object.keys(node.children)
        .filter(function (childId) { return isNodeFocusable(node.children[childId]) || node.children[childId].children; })
        .map(function (childId) { return node.children[childId].index; });
    if (indexes.length <= 0) {
        return undefined;
    }
    return findChildWithIndex(node, closestIndex(indexes, index));
};
/**
 * Inserts given child to the parent's children, keeping indices coherent and compact.
 *
 * @param parent - node to which new child is about to be added
 * @param newChild - node to be added to the parent
 */
var insertChildNode = function (parent, newChild) {
    if (!parent || !newChild) {
        return;
    }
    var childrenIds = Object.keys(parent.children || {});
    var orderedChildren = {};
    newChild.parent = parent.id;
    if (typeof newChild.index !== 'number' || newChild.index > childrenIds.length) {
        newChild.index = childrenIds.length;
    }
    for (var _i = 0, childrenIds_1 = childrenIds; _i < childrenIds_1.length; _i++) {
        var childId = childrenIds_1[_i];
        var child = parent.children[childId];
        // Inserting new child, from now on all further children needs to have index value increased by one
        if (child.index >= newChild.index) {
            if (!orderedChildren[newChild.id]) {
                orderedChildren[newChild.id] = newChild;
            }
            child.index += 1;
        }
        orderedChildren[child.id] = child;
    }
    if (!orderedChildren[newChild.id]) {
        orderedChildren[newChild.id] = newChild;
    }
    parent.children = orderedChildren;
};
/**
 * Removes given child from the parent's children, keeping indices coherent and compact.
 *
 * @param parent - node from which children given child is about to be removed
 * @param childId - id of the node to be removed from parent's children
 */
var removeChildNode = function (parent, childId) {
    if (!childId || !parent || !parent.children || !parent.children[childId]) {
        return;
    }
    var removedChildIndex = parent.children[childId].index;
    delete parent.children[childId];
    for (var _i = 0, _a = Object.keys(parent.children); _i < _a.length; _i++) {
        var childId_1 = _a[_i];
        var child = parent.children[childId_1];
        if (child.index > removedChildIndex) {
            child.index -= 1;
        }
    }
};
/**
 * Returns flatten tree of nodes.
 *
 * Returned node values are kept coherent and delivers basic information about the node:
 *   - missing 'id' and 'parent' fields are filled in
 *   - children subtree is removed
 *
 * E.g.
 *   {
 *      'root': {
 *         children: {
 *           'a': { isFocusable: true },
 *           'b': {
 *             orientation: 'horizontal',
 *             children: {
 *               ba: { isFocusable: true },
 *               bb: { isFocusable: true }
 *           }
 *         }
 *       }
 *     }
 *   }
 *
 * Expect:
 *    {
 *      'root': { id: 'root' }
 *      'a': { id: 'a', parent: 'root', isFocusable: true }
 *      'b': { id: 'b', parent: 'root', orientation: 'horizontal' }
 *      'ba': { id: 'ba', parent: 'b', isFocusable: true }
 *      'bb': { id: 'bb', parent: 'b', isFocusable: true }
 *   }
 *
 * @param tree - tree to flatten
 */
var flattenNodeTree = function (tree) {
    var flatNodeTree = {};
    if (!tree) {
        return flatNodeTree;
    }
    var _flatten = function (tree, parent) {
        Object.keys(tree).forEach(function (nodeId) {
            flatNodeTree[nodeId] = __assign(__assign({}, tree[nodeId]), { id: nodeId, parent: tree[nodeId].parent || parent, children: undefined });
            if (tree[nodeId].children) {
                _flatten(tree[nodeId].children, nodeId);
            }
        });
    };
    _flatten(tree);
    return flatNodeTree;
};
/**
 * Returns flatten node's sub-tree including the given node.
 *
 * Returned node values are kept coherent and delivers basic information about the node:
 *   - missing 'id' and 'parent' fields are filled in
 *   - children subtree is removed
 *
 * E.g.
 *    {
 *      id: 'root'
 *      children: {
 *        'a': { isFocusable: true },
 *        'b': {
 *          orientation: 'horizontal',
 *          children: {
 *            ba: { isFocusable: true },
 *            bb: { isFocusable: true }
 *          }
 *        }
 *      }
 *    }
 *
 * Expect:
 *    {
 *      'root': { id: 'root' }
 *      'a': { id: 'a', parent: 'root', isFocusable: true }
 *      'b': { id: 'b', parent: 'root', orientation: 'horizontal' }
 *      'ba': { id: 'ba', parent: 'b', isFocusable: true }
 *      'bb': { id: 'bb', parent: 'b', isFocusable: true }
 *   }
 *
 * @param node - node which sub-tree is about to be flatten
 */
var flattenNode = function (node) {
    var _a;
    if (!node) {
        return {};
    }
    return flattenNodeTree((_a = {}, _a[node.id] = node, _a));
};
/**
 * Returns valid orientation parameter.
 *
 * TypeScript defines type appropriately, but in JavaScript anything can be passed. Method doesn't imply that user
 * might be malicious, it just tries to unify provided string values, by making them a valid Orientation type.
 *
 * @param orientation - orientation to correct
 */
var toValidOrientation = function (orientation) {
    if (!orientation) {
        return undefined;
    }
    for (var _i = 0, _a = Object.keys(exports.Orientations); _i < _a.length; _i++) {
        var orientationKey = _a[_i];
        if (exports.Orientations[orientationKey] === orientation.toLowerCase()) {
            return exports.Orientations[orientationKey];
        }
    }
    return undefined;
};
/**
 * Returns valid direction parameter.
 *
 * TypeScript defines type appropriately, but in JavaScript anything can be passed. Method doesn't imply that user
 * might be malicious, it just tries to unify provided string values, by making them a valid Orientation type.
 *
 * @param direction - direction to correct
 */
var toValidDirection = function (direction) {
    if (!direction) {
        return undefined;
    }
    for (var _i = 0, _a = Object.keys(exports.Directions); _i < _a.length; _i++) {
        var directionKey = _a[_i];
        if (exports.Directions[directionKey] === direction.toLowerCase()) {
            return exports.Directions[directionKey];
        }
    }
    return undefined;
};

//      
// An event handler can take an optional event argument
// and should not return a value
                                          
                                                               

// An array of all currently registered event handlers for a type
                                            
                                                            
// A map of event types and their corresponding event handlers.
                        
                                 
                                   
  

/** Mitt: Tiny (~200b) functional event emitter / pubsub.
 *  @name mitt
 *  @returns {Mitt}
 */
function mitt(all                 ) {
	all = all || Object.create(null);

	return {
		/**
		 * Register an event handler for the given type.
		 *
		 * @param  {String} type	Type of event to listen for, or `"*"` for all events
		 * @param  {Function} handler Function to call in response to given event
		 * @memberOf mitt
		 */
		on: function on(type        , handler              ) {
			(all[type] || (all[type] = [])).push(handler);
		},

		/**
		 * Remove an event handler for the given type.
		 *
		 * @param  {String} type	Type of event to unregister `handler` from, or `"*"`
		 * @param  {Function} handler Handler function to remove
		 * @memberOf mitt
		 */
		off: function off(type        , handler              ) {
			if (all[type]) {
				all[type].splice(all[type].indexOf(handler) >>> 0, 1);
			}
		},

		/**
		 * Invoke all handlers for the given type.
		 * If present, `"*"` handlers are invoked after type-matched handlers.
		 *
		 * @param {String} type  The event type to invoke
		 * @param {Any} [evt]  Any value (object is recommended and powerful), passed to each handler
		 * @memberOf mitt
		 */
		emit: function emit(type        , evt     ) {
			(all[type] || []).slice().map(function (handler) { handler(evt); });
			(all['*'] || []).slice().map(function (handler) { handler(type, evt); });
		}
	};
}

var Lrud = /** @class */ (function () {
    function Lrud() {
        this.nodes = {};
        this.rootNodeId = undefined;
        this.currentFocusNodeId = undefined;
        this.isIndexAlignMode = false;
        this.emitter = mitt();
        this.overrides = {};
    }
    /**
     * Registers a callback for an LRUD event.
     *
     * @param {string} eventName - event to subscribe to
     * @param {function} callback - function to call on event
     */
    Lrud.prototype.on = function (eventName, callback) {
        this.emitter.on(eventName, callback);
    };
    /**
     * Unregisters a callback for an LRUD event.
     *
     * @param {string} eventName - event to unsubscribe from
     * @param {function} callback - function that was added using .on()
     */
    Lrud.prototype.off = function (eventName, callback) {
        this.emitter.off(eventName, callback);
    };
    /**
     * Returns the root node.
     */
    Lrud.prototype.getRootNode = function () {
        var node = this.getNode(this.rootNodeId);
        if (!node) {
            throw new Error('no root node');
        }
        return node;
    };
    /**
     * Returns the current focused node.
     */
    Lrud.prototype.getCurrentFocusNode = function () {
        return this.getNode(this.currentFocusNodeId);
    };
    /**
     * Registers a new node into the LRUD tree.
     *
     * @param {string} nodeId - id of the node to register
     * @param {object} node - registered node parameters
     * @param {string} [node.id] - if null, `nodeId` is used
     * @param {string} [node.parent] - if null, value of `this.rootNodeId` is used
     * @param {number} [node.index] - if null, index is 1 more than the index of the last sibling. if no previous siblings, index is 1
     * @param {number[]} [node.indexRange] - defaults to null. acts as a colspan, value [0] is lower bound, value [1] is upper bound
     * @param {object} [node.selectAction] - if a node has a selectAction, it is focusable
     * @param {boolean} [node.isFocusable] - if a node is explicitly set as isFocusable, it is focusable
     * @param {boolean} [node.isWrapping] - if true, when asking for the next child at the end or start of the node, the will "wrap around" and return the first/last (when asking for the last/first)
     * @param {string} [node.orientation] - can be "vertical" or "horizontal". is used in conjunction when handling direction of key press, to determine which child is "next"
     * @param {boolean} [node.isIndexAlign]  -if a node is index aligned, its descendents should jump to nodes based on index instead of activeChild
     * @param {function} [node.onLeave] - if a node has an `onLeave` function, it will be run when a move event leaves this node
     * @param {function} [node.onEnter] - if a node has an `onEnter` function, it will be run when a move event enters this node
     */
    Lrud.prototype.registerNode = function (nodeId, node) {
        if (node === void 0) { node = {}; }
        if (this.getNode(nodeId)) {
            throw Error("Node with an ID of " + nodeId + " has already been registered");
        }
        // It is not allowed to register node directly with children, for such purposes registerTree should be used
        node = __assign(__assign({}, node), { id: nodeId });
        delete node.children;
        // if this is the very first node, set it as root and return...
        if (Object.keys(this.nodes).length <= 0) {
            this.rootNodeId = nodeId;
            this.nodes[nodeId] = node;
            return this;
        }
        // if this node has no parent, assume its parent is root
        if (!node.parent && nodeId !== this.rootNodeId) {
            node.parent = this.rootNodeId;
        }
        var parentNode = this.getNode(node.parent);
        // to keep tree coherent, nodes that are about to be added to not existing parent are ignored
        if (!parentNode) {
            return this;
        }
        // add the node into the tree
        this.nodes[nodeId] = node;
        insertChildNode(parentNode, node);
        return this;
    };
    /**
     * Allows to change node's parent by moving it's whole sub-tree.
     *
     * @param {string} nodeId - id of the node that is about to change parent
     * @param {string} newParentNodeId - id of the node that became a new parent for nodeId
     * @param {object} [options]
     * @param {number} [options.index] - index at which nodeId should be inserted as a child of newParentNodeId
     * @param {boolean} [options.maintainIndex] - applies only if index is not defined; if true, node will be inserted at
     *                                           it's current position if possible; otherwise node will be appended; default: false
     */
    Lrud.prototype.moveNode = function (nodeId, newParentNodeId, options) {
        if (options === void 0) { options = { maintainIndex: false }; }
        // It's not possible to move root node
        if (nodeId === this.rootNodeId) {
            return;
        }
        var node = this.getNode(nodeId);
        if (!node) {
            return;
        }
        var newParentNode = this.getNode(newParentNodeId);
        if (!newParentNode) {
            return;
        }
        // There's no need to change the parent
        if (node.parent === newParentNodeId) {
            return;
        }
        var oldParentNode = this.getNode(node.parent);
        // Removing node from old parent
        removeChildNode(oldParentNode, nodeId);
        // Changing parent of a node
        if (typeof options.index === 'number') {
            node.index = options.index;
        }
        else if (!options.maintainIndex) {
            node.index = undefined;
        }
        insertChildNode(newParentNode, node);
        // If moved node was an active child of the old parent, it needs to be cleaned out as well
        this.unsetActiveChild(oldParentNode.id, node.id);
        // If moved node which is (or it's subtree contains) currently focused node, then parent's active child needs to be adjusted
        if (this.isSameOrParentForChild(nodeId, this.currentFocusNodeId)) {
            this.setActiveChildRecursive(newParentNode.id, node.id);
        }
    };
    /**
     * Registers a new node into the LRUD tree.
     *
     * Kept for backwards compatibility reasons.
     *
     * @param {string} nodeId - id of the node to register
     * @param {object} [node] - registered node parameters
     */
    Lrud.prototype.register = function (nodeId, node) {
        return this.registerNode(nodeId, node);
    };
    /**
     * Unregisters a node from the navigation tree.
     *
     * Kept for backwards compatibility reasons.
     *
     * @param {string} nodeId - id of the node to unregister
     * @param {object} [unregisterOptions]
     */
    Lrud.prototype.unregister = function (nodeId, unregisterOptions) {
        this.unregisterNode(nodeId, unregisterOptions);
    };
    /**
     * Unregisters a node from the navigation tree.
     *
     * @param {string} nodeId - id of the node to unregister
     * @param {object} [unregisterOptions]
     * @param {boolean} [unregisterOptions.forceRefocus] if true, LRUD will attempt to re-focus on a new node if the currently focused
     *                                                   node becomes unregistered due to the given node ID being unregistered
     */
    Lrud.prototype.unregisterNode = function (nodeId, unregisterOptions) {
        var _this = this;
        if (unregisterOptions === void 0) { unregisterOptions = { forceRefocus: true }; }
        if (nodeId === this.rootNodeId) {
            this.nodes = {};
            this.rootNodeId = undefined;
            this.currentFocusNodeId = undefined;
            this.isIndexAlignMode = false;
            this.emitter = mitt();
            this.overrides = {};
            return this;
        }
        var node = this.getNode(nodeId);
        // if we're trying to unregister a node that doesn't exist, exit out
        if (!node) {
            return this;
        }
        // get a copy of the node to pass to the blur event, and grab the parent to work with it
        var parentNode = this.getNode(node.parent);
        // delete the node itself (delete from the parent and re-set the parent later)
        removeChildNode(parentNode, nodeId);
        // delete all node's children
        var flatNodeTree = flattenNode(node);
        Object.keys(flatNodeTree).forEach(function (flatNodeId) {
            delete _this.nodes[flatNodeId];
        });
        // ...if we're unregistering the activeChild of our parent (could be a leaf OR branch)
        // we might need to recalculate the focus...
        if (parentNode.activeChild && parentNode.activeChild === nodeId) {
            this.isIndexAlignMode = false;
            // check if the current focus node was removed, if so focus needs to be recalculated
            if (!this.getCurrentFocusNode()) {
                this.currentFocusNodeId = undefined;
                if (unregisterOptions.forceRefocus) {
                    this.recalculateFocus(parentNode);
                }
            }
            this.unsetActiveChild(parentNode.id, nodeId);
        }
        // blur on the nodeClone
        this.emitter.emit('blur', node);
        if (node.onBlur) {
            node.onBlur(node);
        }
        // if we have any overrides whose target or ID is the node (or one of its children) we just unregistered, we should unregister
        // those overrides (thus keeping state clean)
        Object.keys(this.overrides).forEach(function (overrideId) {
            var override = _this.overrides[overrideId];
            if (flatNodeTree[override.target] || flatNodeTree[override.id]) {
                _this.unregisterOverride(overrideId);
            }
        });
        return this;
    };
    /**
     * Registers a new override onto the LRUD instance.
     *
     * @param {string} overrideId - id of the override to register
     * @param {object} override
     * @param {string} override.id - id of the node, for which this override should be triggered
     * @param {string} override.direction - traversal direction, for which this override should be triggered
     * @param {string} override.target - id of the node to which this overrides points
     */
    Lrud.prototype.registerOverride = function (overrideId, override) {
        if (!overrideId) {
            throw new Error('need an ID to register an override');
        }
        if (this.overrides[overrideId]) {
            throw new Error("override with ID of " + overrideId + " already exists");
        }
        if (!override.id) {
            throw new Error("registering override: " + overrideId + " - missing internal id");
        }
        if (!override.direction) {
            throw new Error("registering override: " + overrideId + " - missing internal direction");
        }
        if (!override.target) {
            throw new Error("registering override: " + overrideId + " - missing internal target");
        }
        this.overrides[overrideId] = override;
        return this;
    };
    /**
     * Unregisters an override from the LRUD instance.
     *
     * @param {string} overrideId - id of the override to unregister
     */
    Lrud.prototype.unregisterOverride = function (overrideId) {
        delete this.overrides[overrideId];
        return this;
    };
    /**
     * Returns a node for an ID.
     *
     * @param {string} nodeId - id of the node to be returned
     */
    Lrud.prototype.getNode = function (nodeId) {
        if (!nodeId) {
            return undefined;
        }
        return this.nodes[nodeId];
    };
    /**
     * Gets a node by ID and then unregister it from the instance.
     *
     * @param {string} nodeId - id of the node to be picked
     */
    Lrud.prototype.pickNode = function (nodeId) {
        var node = this.getNode(nodeId);
        if (!node) {
            return undefined;
        }
        this.unregisterNode(nodeId);
        return node;
    };
    /**
     * Starting from a node, climb up the navigation tree until we find a node that can be
     * actioned, based on the given direction. An actionable node is one whose orientation is valid
     * for the given direction, has focusable children and whose activeChild isn't a leaf that is
     * also its current activeChild.
     *
     * @param {object} node - node from which climbing up starts
     * @param {string} direction - direction in which to traverse while climbing up
     */
    Lrud.prototype.climbUp = function (node, direction) {
        if (!node) {
            return undefined;
        }
        // if we have a matching override at this point in the climb, return that target node
        for (var _i = 0, _a = Object.keys(this.overrides); _i < _a.length; _i++) {
            var overrideId = _a[_i];
            var override = this.overrides[overrideId];
            if (override.id === node.id && override.direction.toUpperCase() === direction.toUpperCase()) {
                return this.getNode(override.target);
            }
        }
        // if we're on a currently focused node, climb up, definitely we are looking for some other node
        if (node.id === this.currentFocusNodeId) {
            return this.climbUp(this.getNode(node.parent), direction);
        }
        // we have children, but the orientation doesn't match, so try our parent
        if (!isDirectionAndOrientationMatching(node.orientation, direction)) {
            return this.climbUp(this.getNode(node.parent), direction);
        }
        // if we couldn't find any focusable candidate within children or we get currently
        // activeChild, we have to look for other focusable candidate, climb up
        if (!this.getNextFocusableChildInDirection(node, direction)) {
            return this.climbUp(this.getNode(node.parent), direction);
        }
        return node;
    };
    /**
     * Starting from the given node, dig down the navigation tree until we find a focusable
     * leaf, and return it. dig "direction" priority:
     * - index align mode
     * - active child
     * - first focusable child
     *
     * @param {object} node - node, from which digging down starts
     * @param {string} [direction] - direction in which to traverse while digging down
     */
    Lrud.prototype.digDown = function (node, direction) {
        if (direction === void 0) { direction = exports.Directions.UNSPECIFIED; }
        if (!node) {
            return undefined;
        }
        // If the node is focusable, then return it, but only when it doesn't contain focusable children.
        // Otherwise, digging down to "the deepest" focusable node.
        // Focusable "leaf" has a higher priority than focusable "container".
        if (isNodeFocusable(node) && (node.isStopPropagate || !this.doesNodeHaveFocusableChildren(node))) {
            return node;
        }
        /*
        if we're in a nested grid
          if we're going VERTICAL DOWN
            take the first child, and then match the index
          if we're going VERTICAL UP
            take the last child, and then match the index
    
        if we're in a nested grid
          and we're going HORIZONTAL LEFT
            take the matching index of the same depth, and then the last child
          and we're going HORIZONTAL RIGHT
            take the matching index of the same depth, and then the first child
    
        if its not a nested grid, take the matching index
        */
        if (this.isIndexAlignMode) {
            var currentFocusedNode = this.getCurrentFocusNode();
            var currentFocusedIndexRange = currentFocusedNode.indexRange;
            var currentFocusedIndex = currentFocusedIndexRange ? currentFocusedIndexRange[0] : currentFocusedNode.index;
            if (node.isIndexAlign) {
                // we're in a nested grid, so need to take into account orientation and direction of travel
                var nodeParent = this.getNode(node.parent);
                if (nodeParent.orientation === exports.Orientations.VERTICAL) {
                    if (direction === exports.Directions.UP) {
                        return this.digDown(findChildWithClosestIndex(this.getNodeLastChild(node), currentFocusedIndex, currentFocusedIndexRange), direction);
                    }
                    if (direction === exports.Directions.DOWN) {
                        return this.digDown(findChildWithClosestIndex(this.getNodeFirstChild(node), currentFocusedIndex, currentFocusedIndexRange), direction);
                    }
                }
                if (nodeParent.orientation === exports.Orientations.HORIZONTAL) {
                    if (direction === exports.Directions.LEFT) {
                        var firstStep = findChildWithClosestIndex(node, this.getNode(currentFocusedNode.parent).index);
                        return this.digDown(this.getNodeLastChild(firstStep), direction);
                    }
                    if (direction === exports.Directions.RIGHT) {
                        var firstStep = findChildWithClosestIndex(node, this.getNode(currentFocusedNode.parent).index);
                        return this.digDown(this.getNodeFirstChild(firstStep), direction);
                    }
                }
            }
            // we're not in a nested grid, so just look for matching index ranges or index
            var matchingViaIndexRange = findChildWithMatchingIndexRange(node, currentFocusedIndex);
            if (matchingViaIndexRange) {
                return this.digDown(matchingViaIndexRange, direction);
            }
            return this.digDown(findChildWithClosestIndex(node, currentFocusedIndex, currentFocusedIndexRange), direction);
        }
        // if possible, picking a branch that had focus in the past, one of its children was focused
        if (node.activeChild) {
            return this.digDown(node.children[node.activeChild], direction);
        }
        // otherwise simply digging deeper, picking branch with first focusable candidate
        return this.digDown(this.getNextFocusableChildInDirection(node, exports.Directions.UNSPECIFIED), direction);
    };
    /**
     * Gets the semantic next focusable child for a given direction.
     *
     * If the direction is 'left' or 'up', return the semantic previous focusable child of the node.
     * If the direction is 'right' or 'down'', return the semantic next focusable child of the node.
     * If the direction is *, return the semantic next (or previous, if next not found) focusable child of the node.
     *
     * @param {object} node - node, for which next focusable child for a given direction is returned
     * @param {string} direction - direction in which to traverse while searching for next focusable child
     */
    Lrud.prototype.getNextFocusableChildInDirection = function (node, direction) {
        if (!node) {
            return undefined;
        }
        var validOrientation = toValidOrientation(node.orientation);
        var validDirection = toValidDirection(direction);
        var nextChildInDirection;
        var traverseForward = (validDirection === exports.Directions.UNSPECIFIED) ||
            (validOrientation === exports.Orientations.HORIZONTAL && validDirection === exports.Directions.RIGHT) ||
            (validOrientation === exports.Orientations.VERTICAL && validDirection === exports.Directions.DOWN);
        if (traverseForward) {
            nextChildInDirection = this.getNextFocusableChild(node);
        }
        var traverseBackward = (validDirection === exports.Directions.UNSPECIFIED) ||
            (validOrientation === exports.Orientations.HORIZONTAL && validDirection === exports.Directions.LEFT) ||
            (validOrientation === exports.Orientations.VERTICAL && validDirection === exports.Directions.UP);
        if (!nextChildInDirection && traverseBackward) {
            nextChildInDirection = this.getPrevFocusableChild(node);
        }
        return nextChildInDirection;
    };
    /**
     * Gets the semantic "next" child for a node that might be focused or bypass focus to its children.
     *
     * @param {object} node - node, for which next focusable child is returned
     */
    Lrud.prototype.getNextFocusableChild = function (node) {
        if (!node || !node.children) {
            return undefined;
        }
        // there's no child that is (or was) focused, so we can quickly pick first focusable child
        var activeChild = this.getNode(node.activeChild);
        if (!activeChild) {
            return this.getNodeFirstFocusableChild(node);
        }
        var nextChild = activeChild;
        do {
            nextChild = findChildWithIndex(node, nextChild.index + 1);
            if (!nextChild && node.isWrapping) {
                nextChild = this.getNodeFirstFocusableChild(node);
            }
            if (!nextChild) {
                nextChild = activeChild;
            }
        } while (!this.isNodeFocusableCandidate(nextChild) && nextChild.id !== activeChild.id);
        return nextChild.id !== activeChild.id ? nextChild : undefined;
    };
    /**
     * Gets the semantic "previous" child for a node that might be focused or bypass focus to its children.
     *
     * @param {object} node - node, for which previous focusable child is returned
     */
    Lrud.prototype.getPrevFocusableChild = function (node) {
        if (!node || !node.children) {
            return undefined;
        }
        // there's no child that is (or was) focused, so we can quickly pick last focusable child
        var activeChild = this.getNode(node.activeChild);
        if (!activeChild) {
            return this.getNodeLastFocusableChild(node);
        }
        // starting from child that is (or was) focused
        var prevChild = activeChild;
        do {
            prevChild = findChildWithIndex(node, prevChild.index - 1);
            if (!prevChild && node.isWrapping) {
                prevChild = this.getNodeLastFocusableChild(node);
            }
            if (!prevChild) {
                prevChild = activeChild;
            }
        } while (!this.isNodeFocusableCandidate(prevChild) && prevChild.id !== activeChild.id);
        return prevChild.id !== activeChild.id ? prevChild : undefined;
    };
    /**
     * Gets the first child of a node, based on index.
     *
     * @param {object} node - node, for which first child is returned
     */
    Lrud.prototype.getNodeFirstChild = function (node) {
        if (!node || !node.children) {
            return undefined;
        }
        return findChildWithIndex(node, 0);
    };
    /**
     * Gets the last child of a node, based on index.
     *
     * @param {object} node - node, for which last child is returned
     */
    Lrud.prototype.getNodeLastChild = function (node) {
        if (!node || !node.children) {
            return undefined;
        }
        return findChildWithIndex(node, Object.keys(node.children).length - 1);
    };
    /**
     * Gets the first focusable (or containing focusable nodes) child of a node, based on index.
     *
     * @param {object} node - node, for which first focusable child is returned
     */
    Lrud.prototype.getNodeFirstFocusableChild = function (node) {
        if (!node || !node.children) {
            return undefined;
        }
        var orderedChildren = Object.keys(node.children)
            .map(function (childId) { return node.children[childId]; })
            .sort(function (a, b) { return a.index - b.index; });
        for (var _i = 0, orderedChildren_1 = orderedChildren; _i < orderedChildren_1.length; _i++) {
            var child = orderedChildren_1[_i];
            if (this.isNodeFocusableCandidate(child)) {
                return child;
            }
        }
        return undefined;
    };
    /**
     * Gets the last focusable (or containing focusable nodes) child of a node, based on index.
     *
     * @param {object} node - node, for which last focusable child is returned
     */
    Lrud.prototype.getNodeLastFocusableChild = function (node) {
        if (!node || !node.children) {
            return undefined;
        }
        var inverselyOrderedChildren = Object.keys(node.children)
            .map(function (childId) { return node.children[childId]; })
            .sort(function (a, b) { return b.index - a.index; });
        for (var _i = 0, inverselyOrderedChildren_1 = inverselyOrderedChildren; _i < inverselyOrderedChildren_1.length; _i++) {
            var child = inverselyOrderedChildren_1[_i];
            if (this.isNodeFocusableCandidate(child)) {
                return child;
            }
        }
        return undefined;
    };
    /**
     * Given an event, handle any state changes that may arise from the direction pressed.
     *
     * State changes based on climbing up and digging down from the current focusedNode
     *
     * @param {object} event
     * @param {number} [event.keyCode]
     * @param {string} [event.direction]
     * @param {object} [options]
     * @param {boolean} [options.forceFocus] - if true and there's no currently focused node, LRUD will try to focus
     *                                         first focusable node; default: false
     */
    Lrud.prototype.handleKeyEvent = function (event, options) {
        if (options === void 0) { options = { forceFocus: false }; }
        if (!event) {
            return undefined;
        }
        var direction = (event.keyCode) ? getDirectionForKeyCode(event.keyCode) : toValidDirection(event.direction);
        if (!direction) {
            return undefined;
        }
        var currentFocusNode = this.getCurrentFocusNode();
        // if all we're doing is processing an enter, just run the `onSelect` function of the current node...
        if (direction === exports.Directions.ENTER) {
            if (currentFocusNode) {
                this.emitter.emit('select', currentFocusNode);
                if (currentFocusNode.onSelect) {
                    currentFocusNode.onSelect(currentFocusNode);
                }
            }
            return currentFocusNode;
        }
        var topNode;
        var focusableNode;
        if (!currentFocusNode && options.forceFocus) {
            // No node is focused, focusing first focusable node
            topNode = this.getRootNode();
            focusableNode = this.getNextFocusableChildInDirection(topNode, exports.Directions.UNSPECIFIED);
        }
        else {
            // climb up from where we are...
            topNode = this.climbUp(currentFocusNode, direction);
            // ... if we cant find a top node, its an invalid move - just return
            if (!topNode) {
                return undefined;
            }
            // ...if we need to align indexes, turn the flag on now...
            this.isIndexAlignMode = topNode.isIndexAlign === true;
            // ...get the top's next child in the direction we're going...
            var nextChild = this.getNextFocusableChildInDirection(topNode, direction);
            // ...and depending on if we're able to find a child, dig down from the child or from the original top...
            focusableNode = this.digDown(nextChild || topNode, direction);
        }
        if (!focusableNode) {
            return undefined;
        }
        // ...give an opportunity for the move to be cancelled by the leaving node
        if (currentFocusNode && currentFocusNode.shouldCancelLeave) {
            if (currentFocusNode.shouldCancelLeave(currentFocusNode, focusableNode)) {
                if (currentFocusNode.onLeaveCancelled) {
                    currentFocusNode.onLeaveCancelled(currentFocusNode, focusableNode);
                }
                this.emitter.emit('cancelled', {
                    leave: currentFocusNode,
                    enter: focusableNode
                });
                return currentFocusNode;
            }
        }
        // ...give an opportunity for the move to be cancelled by the entering node
        if (focusableNode.shouldCancelEnter) {
            if (focusableNode.shouldCancelEnter(currentFocusNode, focusableNode)) {
                if (focusableNode.onEnterCancelled) {
                    focusableNode.onEnterCancelled(currentFocusNode, focusableNode);
                }
                this.emitter.emit('cancelled', {
                    leave: currentFocusNode,
                    enter: focusableNode
                });
                return currentFocusNode;
            }
        }
        // ...and then assign focus
        this.assignFocus(focusableNode.id);
        var offset = (direction === exports.Directions.LEFT || direction === exports.Directions.UP) ? -1 : 1;
        // emit events and fire functions now that the move has completed
        this.emitter.emit('move', {
            leave: currentFocusNode,
            enter: focusableNode,
            direction: direction,
            offset: offset
        });
        if (topNode.onMove) {
            topNode.onMove({
                node: topNode,
                leave: currentFocusNode,
                enter: focusableNode,
                direction: direction,
                offset: offset
            });
        }
        if (currentFocusNode && currentFocusNode.onLeave) {
            currentFocusNode.onLeave(currentFocusNode);
        }
        if (focusableNode.onEnter) {
            focusableNode.onEnter(focusableNode);
        }
        return focusableNode;
    };
    /**
     * Sets the activeChild of the parentId node to the value of the childId node.
     *
     * @param {string} parentId - id of the node, which activeChild is about to be set
     * @param {string} childId - id of the node, that is about to be set as parent's activeChild
     */
    Lrud.prototype.setActiveChild = function (parentId, childId) {
        var child = this.getNode(childId);
        var parent = this.getNode(parentId);
        if (!parent || !child) {
            return;
        }
        if (child.parent !== parent.id) {
            return;
        }
        // the parent already has an active child, and its NOT the same child that we're now setting
        if (parent.activeChild && parent.activeChild !== child.id) {
            var currentActiveChild = this.getNode(parent.activeChild);
            parent.activeChild = child.id;
            if (currentActiveChild) {
                this.emitter.emit('inactive', currentActiveChild);
                if (currentActiveChild.onInactive) {
                    currentActiveChild.onInactive(currentActiveChild);
                }
            }
            this.emitter.emit('active', child);
            if (child.onActive) {
                child.onActive(child);
            }
            if (parent.onActiveChildChange) {
                parent.onActiveChildChange({
                    node: parent,
                    leave: currentActiveChild,
                    enter: child
                });
            }
        }
        else if (!parent.activeChild) {
            parent.activeChild = child.id;
            this.emitter.emit('active', child);
            if (child.onActive) {
                child.onActive(child);
            }
            if (parent.onActiveChildChange) {
                parent.onActiveChildChange({
                    node: parent,
                    leave: null,
                    enter: child
                });
            }
        }
    };
    /**
     * Sets the activeChild of the parentId node to the value of the childId node.
     *
     * If the parent node has a parent itself, it digs up the tree and sets those activeChild values.
     *
     * @param {string} parentId - id of the node, which activeChild is about to be set
     * @param {string} childId - id of the node, that is about to be set as parent's activeChild
     */
    Lrud.prototype.setActiveChildRecursive = function (parentId, childId) {
        this.setActiveChild(parentId, childId);
        var parent = this.getNode(parentId);
        // if the parent has a parent, bubble up
        if (parent && parent.parent) {
            this.setActiveChildRecursive(parent.parent, parent.id);
        }
    };
    /**
     * Unsets the activeChild of the parent nodes branch ensuring that activeChild is on the unsetting childId path
     * and not on the currentFocusNode's path, unless childId is a currentFocusNode.
     *
     * @param {string} parentId - id of the node, which activeChild is about to be unset
     * @param {string} activeChildId - id of the node, that is about to be unset as parent's activeChild
     */
    Lrud.prototype.unsetActiveChild = function (parentId, activeChildId) {
        var parent = this.getNode(parentId);
        if (!parent || !parent.activeChild) {
            return;
        }
        if (parent.activeChild !== activeChildId) {
            return;
        }
        var isActiveChildAtCurrentFocusNodeBranch = this.isSameOrParentForChild(activeChildId, this.currentFocusNodeId);
        while (parent && parent.activeChild) {
            var isParentAtActiveChildBranch = parent.activeChild === activeChildId;
            var isParentAtCurrentFocusedNodeBranch = this.isSameOrParentForChild(parent.activeChild, this.currentFocusNodeId);
            if (isActiveChildAtCurrentFocusNodeBranch || (isParentAtActiveChildBranch && !isParentAtCurrentFocusedNodeBranch)) {
                delete parent.activeChild;
            }
            activeChildId = parent.id;
            parent = this.getNode(parent.parent);
        }
    };
    /**
     * Sets the current focus of the instance to the given node ID.
     *
     * If the given node ID points to a non-focusable node, we dig down from
     * the given node to find a node that can be focused on.
     *
     * Calls `onFocus` on the given node, if it exists, and emits a `focus` event,
     * also calls `onBlur` on the node that WAS focused before this function was called.
     *
     * @param {string} nodeId - id of the node to be focused
     */
    Lrud.prototype.assignFocus = function (nodeId) {
        var node = this.getNode(nodeId);
        // Focus might be assigned to node that is not focusable itself, but
        // contains focusable children, looking for such child
        if (node && !isNodeFocusable(node)) {
            node = this.digDown(node, exports.Directions.UNSPECIFIED);
        }
        if (!node) {
            throw new Error('trying to assign focus to a non focusable node');
        }
        var previouslyFocusedNode = this.getCurrentFocusNode();
        if (previouslyFocusedNode) {
            this.emitter.emit('blur', previouslyFocusedNode);
            if (previouslyFocusedNode.onBlur) {
                previouslyFocusedNode.onBlur(previouslyFocusedNode);
            }
        }
        this.currentFocusNodeId = node.id;
        if (node.parent) {
            this.setActiveChildRecursive(node.parent, node.id);
        }
        if (node.onFocus) {
            node.onFocus(node);
        }
        this.emitter.emit('focus', node);
    };
    /**
     * If the focus of the tree is out of sync, ie, the current focused node becomes unfocusable,
     * this can be used to fall back to another focus.
     *
     * @param {object} node - node, based on which focus is recalculated
     */
    Lrud.prototype.recalculateFocus = function (node) {
        var topNode = this.climbUp(node, exports.Directions.UNSPECIFIED) || this.getRootNode();
        var nextChild = this.getNextFocusableChildInDirection(topNode, exports.Directions.UNSPECIFIED);
        var focusableNode = this.digDown(nextChild || topNode, exports.Directions.UNSPECIFIED);
        if (focusableNode) {
            this.assignFocus(focusableNode.id);
        }
        else {
            this.currentFocusNodeId = undefined;
        }
    };
    /**
     * Given a tree, register all of its nodes into this instance.
     *
     * @param {object} tree
     */
    Lrud.prototype.registerTree = function (tree) {
        var _this = this;
        var flatNodeTree = flattenNodeTree(tree);
        Object.keys(flatNodeTree).forEach(function (flatNodeId) {
            _this.registerNode(flatNodeId, flatNodeTree[flatNodeId]);
        });
    };
    /**
     * Given a tree object, attempt to register that tree into the current lrud instance.
     *
     * If the given tree already exists as a branch in the instance tree, the new tree will replace that branch.
     *
     * If the new tree doesn't already exist as a branch in the instance tree, this function will register the new
     * tree as a branch against the root node, as per standard registerNode() behaviour.
     *
     * @param {object} tree
     * @param {object} [options]
     * @param {boolean} [options.maintainIndex] - if true, and new tree is replacing an existing branch of the tree,
     *                                            maintain the original branches relative index; default: true
     */
    Lrud.prototype.insertTree = function (tree, options) {
        if (options === void 0) { options = { maintainIndex: true }; }
        if (!tree) {
            return;
        }
        var replacementNode = tree[Object.keys(tree)[0]];
        replacementNode.id = Object.keys(tree)[0];
        var originalNode = this.pickNode(replacementNode.id);
        if (!replacementNode.parent && originalNode && originalNode.parent) {
            replacementNode.parent = originalNode.parent;
        }
        if (options.maintainIndex && originalNode && typeof originalNode.index === 'number') {
            replacementNode.index = originalNode.index;
        }
        this.registerTree(tree);
    };
    /**
     * Checks is node contains children that might be focused (are a focusable candidates).
     * It checks the whole node's children sub-tree, not only direct children.
     *
     * @param node - node, which children are checked against being focusable candidates
     */
    Lrud.prototype.doesNodeHaveFocusableChildren = function (node) {
        if (!node || !node.children) {
            return false;
        }
        var flatNodeChildren = flattenNodeTree(node.children);
        for (var _i = 0, _a = Object.keys(flatNodeChildren); _i < _a.length; _i++) {
            var flatNodeId = _a[_i];
            if (isNodeFocusable(flatNodeChildren[flatNodeId])) {
                return true;
            }
        }
        return false;
    };
    /**
     * Checks if node is focusable or contains focusable children.
     *
     * @param {object} node - node to check against being focusable candidate
     */
    Lrud.prototype.isNodeFocusableCandidate = function (node) {
        return isNodeFocusable(node) || this.doesNodeHaveFocusableChildren(node);
    };
    /**
     * Checks if given parent node is a child's parent node or the node itself. If it's a parent it doesn't have
     * to be a direct one, but has to be placed at a path to the root node.
     *
     * E.g.
     *        root
     *        / \
     *       A   B
     *      /     \
     *     AA      BA
     *    /  \
     *  AAA  AAB
     *
     *  Expect:
     *    isSameOrParentForChild('A', 'A') -> true
     *    isSameOrParentForChild('A', 'AA') -> true
     *    isSameOrParentForChild('AA', 'A') -> false
     *    isSameOrParentForChild('A', 'AAA') -> true
     *    isSameOrParentForChild('AAA', 'AA') -> false
     *    isSameOrParentForChild('A', 'BA') -> false
     *    isSameOrParentForChild('BA', 'A') -> false
     *    isSameOrParentForChild('AA', 'BA') -> false
     *    isSameOrParentForChild('BA', 'AA') -> false
     *
     * @param {string} parentId - id of the examined parent
     * @param {string} childId - id of the node which parents are queried
     */
    Lrud.prototype.isSameOrParentForChild = function (parentId, childId) {
        if (!parentId || !childId) {
            return false;
        }
        if (parentId === childId) {
            return true;
        }
        var child = this.getNode(childId);
        while (child) {
            if (child.parent === parentId) {
                return true;
            }
            child = this.getNode(child.parent);
        }
        return false;
    };
    /**
     * Changes the ability of a node to be focused in place.
     *
     * @param {string} nodeId - id of the node, which focusability property is about to be changed
     * @param {boolean} isFocusable - focusability value to set
     */
    Lrud.prototype.setNodeFocusable = function (nodeId, isFocusable) {
        var node = this.getNode(nodeId);
        if (!node) {
            return;
        }
        var nodeIsFocusable = isNodeFocusable(node);
        if (nodeIsFocusable === isFocusable) {
            return;
        }
        node.isFocusable = isFocusable;
        if (!isFocusable) {
            if (this.currentFocusNodeId === nodeId) {
                this.recalculateFocus(node);
            }
            if (node.parent) {
                this.unsetActiveChild(node.parent, nodeId);
            }
        }
    };
    return Lrud;
}());

exports.Lrud = Lrud;
