import { Direction, Node, NodeId, NodeIndex, NodeIndexRange, NodeTree, Orientation } from './interfaces';
/**
 * Given an array of values and a goal, return the value from values which is closest to the goal.
 *
 * @param {number[]} values
 * @param {number} goal
 */
export declare const closestIndex: (values: NodeIndex[], goal: NodeIndex) => number;
/**
 * Checks if a given node is focusable.
 *
 * @param {object} node - node to check against focusability
 */
export declare const isNodeFocusable: (node: Node) => boolean;
/**
 * Given a keyCode, lookup and return the direction from the keycodes mapping file.
 *
 * @param {number} keyCode - key code for which corresponding direction is searched
 */
export declare const getDirectionForKeyCode: (keyCode: number) => Direction | undefined;
/**
 * Given an orientation and a direction, do they match?
 *
 * I.e an orientation `horizontal` and direction `left` or `right` is considered matching.
 *
 * Direction CAN be passed as `*` (wildcard). this will always return true.
 *
 * @param {string} orientation - orientation to match with the direction
 * @param {string} direction - direction to match with the orientation
 */
export declare const isDirectionAndOrientationMatching: (orientation: Orientation, direction: Direction) => boolean;
/**
 * Returns a child from the given node whose indexRange encompasses the given index.
 *
 * @param {object} node - node, which children index ranges will be examined
 * @param {number} index - examined index value
 */
export declare const findChildWithMatchingIndexRange: (node: Node, index: NodeIndex) => Node | undefined;
/**
 * Returns a child from the given node whose index matches the given index.
 *
 * @param {object} node - node, which children indices will be examined
 * @param {number} index - searched index value
 */
export declare const findChildWithIndex: (node: Node, index: NodeIndex) => Node | undefined;
/**
 * Returns a child from the given node whose index is numerically closest to the given index.
 *
 * If an indexRange is provided, first check if the node's activeChild is inside
 * the indexRange. If it is, return the activeChild node instead
 *
 * @param {object} node
 * @param {number} index
 * @param {number[]} indexRange
 */
export declare const findChildWithClosestIndex: (node: Node, index: NodeIndex, indexRange?: NodeIndexRange) => Node | undefined;
/**
 * Inserts given child to the parent's children, keeping indices coherent and compact.
 *
 * @param parent - node to which new child is about to be added
 * @param newChild - node to be added to the parent
 */
export declare const insertChildNode: (parent: Node, newChild: Node) => void;
/**
 * Removes given child from the parent's children, keeping indices coherent and compact.
 *
 * @param parent - node from which children given child is about to be removed
 * @param childId - id of the node to be removed from parent's children
 */
export declare const removeChildNode: (parent: Node, childId: NodeId) => void;
/**
 * Returns flatten tree of nodes.
 *
 * Returned node values are kept coherent and delivers basic information about the node:
 *   - missing 'id' and 'parent' fields are filled in
 *   - children subtree is removed
 *
 * E.g.
 *   {
 *      'root': {
 *         children: {
 *           'a': { isFocusable: true },
 *           'b': {
 *             orientation: 'horizontal',
 *             children: {
 *               ba: { isFocusable: true },
 *               bb: { isFocusable: true }
 *           }
 *         }
 *       }
 *     }
 *   }
 *
 * Expect:
 *    {
 *      'root': { id: 'root' }
 *      'a': { id: 'a', parent: 'root', isFocusable: true }
 *      'b': { id: 'b', parent: 'root', orientation: 'horizontal' }
 *      'ba': { id: 'ba', parent: 'b', isFocusable: true }
 *      'bb': { id: 'bb', parent: 'b', isFocusable: true }
 *   }
 *
 * @param tree - tree to flatten
 */
export declare const flattenNodeTree: (tree: NodeTree) => NodeTree;
/**
 * Returns flatten node's sub-tree including the given node.
 *
 * Returned node values are kept coherent and delivers basic information about the node:
 *   - missing 'id' and 'parent' fields are filled in
 *   - children subtree is removed
 *
 * E.g.
 *    {
 *      id: 'root'
 *      children: {
 *        'a': { isFocusable: true },
 *        'b': {
 *          orientation: 'horizontal',
 *          children: {
 *            ba: { isFocusable: true },
 *            bb: { isFocusable: true }
 *          }
 *        }
 *      }
 *    }
 *
 * Expect:
 *    {
 *      'root': { id: 'root' }
 *      'a': { id: 'a', parent: 'root', isFocusable: true }
 *      'b': { id: 'b', parent: 'root', orientation: 'horizontal' }
 *      'ba': { id: 'ba', parent: 'b', isFocusable: true }
 *      'bb': { id: 'bb', parent: 'b', isFocusable: true }
 *   }
 *
 * @param node - node which sub-tree is about to be flatten
 */
export declare const flattenNode: (node: Node) => NodeTree;
/**
 * Returns valid orientation parameter.
 *
 * TypeScript defines type appropriately, but in JavaScript anything can be passed. Method doesn't imply that user
 * might be malicious, it just tries to unify provided string values, by making them a valid Orientation type.
 *
 * @param orientation - orientation to correct
 */
export declare const toValidOrientation: (orientation: Orientation) => Orientation | undefined;
/**
 * Returns valid direction parameter.
 *
 * TypeScript defines type appropriately, but in JavaScript anything can be passed. Method doesn't imply that user
 * might be malicious, it just tries to unify provided string values, by making them a valid Orientation type.
 *
 * @param direction - direction to correct
 */
export declare const toValidDirection: (direction: Direction) => Direction | undefined;
